@@ -4,7 +4,9 @@ import math
 import torch
 from torch import nn, einsum
 import torch.nn.functional as F
+
 from einops import rearrange, repeat, reduce
+from einops.layers.torch import Rearrange
 
 # helper functions
 
@@ -41,6 +43,14 @@ def Downsample(dim, dim_out):
 
 # normalization
 
+class Residual(nn.Module):
+    def __init__(self, fn):
+        super().__init__()
+        self.fn = fn
+
+    def forward(self, x):
+        return self.fn(x) + x
+
 class LayerNorm(nn.Module):
     def __init__(self, dim):
         super().__init__()
@@ -52,18 +62,6 @@ class LayerNorm(nn.Module):
         mean = torch.mean(x, dim = 1, keepdim = True)
         return (x - mean) / (var + eps).sqrt() * self.gamma
 
-class WeightStandardizedConv2d(nn.Conv2d):
-    def forward(self, x):
-        eps = 1e-5 if x.dtype == torch.float32 else 1e-3
-
-        weight = self.weight
-
-        mean = reduce(weight, 'o ... -> o 1 1 1', 'mean')
-        var = reduce(weight, 'o ... -> o 1 1 1', partial(torch.var, unbiased = False))
-        weight = (weight - mean) * (var + eps).rsqrt()
-
-        return F.conv2d(x, weight, self.bias, self.stride, self.padding, self.dilation, self.groups)
-
 class WeightStandardizedConv3d(nn.Conv3d):
     def forward(self, x):
         eps = 1e-5 if x.dtype == torch.float32 else 1e-3
@@ -79,15 +77,23 @@ class WeightStandardizedConv3d(nn.Conv3d):
 # resnet blocks
 
 class Block(nn.Module):
-    def __init__(self, dim, dim_out, groups = 8, weight_standardize = False):
+    def __init__(
+        self,
+        dim,
+        dim_out,
+        groups = 8,
+        weight_standardize = False,
+        frame_kernel_size = 1
+    ):
         super().__init__()
+        kernel_conv_kwargs = partial(kernel_and_same_pad, frame_kernel_size)
         conv = nn.Conv3d if not weight_standardize else WeightStandardizedConv3d
 
-        self.proj = conv(dim, dim_out, (1, 3, 3), padding = (0, 1, 1))
+        self.proj = conv(dim, dim_out, **kernel_conv_kwargs(3, 3))
         self.norm = nn.GroupNorm(groups, dim_out)
         self.act = nn.SiLU()
 
-    def forward(self, x, scale_shift = None):
+    def forward(self, x):
         x = self.proj(x)
         x = self.norm(x)
         return self.act(x)
@@ -98,17 +104,18 @@ class ResnetBlock(nn.Module):
         dim,
         dim_out,
         groups = 8,
+        frame_kernel_size = 1,
         nested_unet_depth = 0,
         nested_unet_dim = 32,
         weight_standardize = False
     ):
         super().__init__()
-        self.block1 = Block(dim, dim_out, groups = groups, weight_standardize = weight_standardize)
+        self.block1 = Block(dim, dim_out, groups = groups, weight_standardize = weight_standardize, frame_kernel_size = frame_kernel_size)
 
         if nested_unet_depth > 0:
-            self.block2 = NestedResidualUnet(dim_out, depth = nested_unet_depth, M = nested_unet_dim, add_residual = True, weight_standardize = weight_standardize)
+            self.block2 = NestedResidualUnet(dim_out, depth = nested_unet_depth, M = nested_unet_dim, frame_kernel_size = frame_kernel_size, weight_standardize = weight_standardize, add_residual = True)
         else:
-            self.block2 = Block(dim_out, dim_out, groups = groups, weight_standardize = weight_standardize)
+            self.block2 = Block(dim_out, dim_out, groups = groups, weight_standardize = weight_standardize, frame_kernel_size = frame_kernel_size)
 
         self.res_conv = nn.Conv3d(dim, dim_out, 1) if dim != dim_out else nn.Identity()
 
@@ -127,17 +134,20 @@ class ConvNextBlock(nn.Module):
         dim_out,
         *,
         mult = 2,
+        frame_kernel_size = 1,
         nested_unet_depth = 0,
         nested_unet_dim = 32
     ):
         super().__init__()
-        self.ds_conv = nn.Conv3d(dim, dim, (1, 7, 7), padding = (0, 3, 3), groups = dim)
+        kernel_conv_kwargs = partial(kernel_and_same_pad, frame_kernel_size)
+
+        self.ds_conv = nn.Conv3d(dim, dim, **kernel_conv_kwargs(7, 7), groups = dim)
 
         self.net = nn.Sequential(
             LayerNorm(dim),
-            nn.Conv3d(dim, dim_out * mult, (1, 3, 3), padding = (0, 1, 1)),
+            nn.Conv3d(dim, dim_out * mult, **kernel_conv_kwargs(3, 3)),
             nn.GELU(),
-            nn.Conv3d(dim_out * mult, dim_out, (1, 3, 3), padding = (0, 1, 1))
+            nn.Conv3d(dim_out * mult, dim_out, **kernel_conv_kwargs(3, 3))
         )
 
         self.nested_unet = NestedResidualUnet(dim_out, depth = nested_unet_depth, M = nested_unet_dim, add_residual = True) if nested_unet_depth > 0 else nn.Identity()
@@ -150,6 +160,18 @@ class ConvNextBlock(nn.Module):
         h = self.nested_unet(h)
         return h + self.res_conv(x)
 
+# feedforward
+
+def FeedForward(dim, mult = 4.):
+    inner_dim = int(dim * mult)
+    return Residual(nn.Sequential(
+        LayerNorm(dim),
+        nn.Conv3d(dim, inner_dim, 1, bias = False),
+        nn.GELU(),
+        LayerNorm(inner_dim),   # properly credit assign normformer
+        nn.Conv3d(inner_dim, dim, 1, bias = False)
+    ))
+
 # attention
 
 class Attention(nn.Module):
@@ -158,7 +180,7 @@ class Attention(nn.Module):
         dim,
         heads = 4,
         dim_head = 64,
-        scale = 16
+        scale = 8
     ):
         super().__init__()
         self.scale = scale
@@ -172,6 +194,8 @@ class Attention(nn.Module):
     def forward(self, x):
         f, h, w = x.shape[-3:]
 
+        residual = x.clone()
+
         x = self.norm(x)
 
         q, k, v = self.to_qkv(x).chunk(3, dim = 1)
@@ -185,7 +209,24 @@ class Attention(nn.Module):
         out = einsum('b h i j, b h j d -> b h i d', attn, v)
 
         out = rearrange(out, 'b h (f x y) d -> b (h d) f x y', f = f, x = h, y = w)
-        return self.to_out(out)
+        return self.to_out(out) + residual
+
+class TransformerBlock(nn.Module):
+    def __init__(
+        self,
+        dim,
+        *,
+        depth,
+        **kwargs
+    ):
+        super().__init__()
+        self.attn = Attention(dim, **kwargs)
+        self.ff =FeedForward(dim)
+
+    def forward(self, x):
+        x = self.attn(x)
+        x = self.ff(x)
+        return x
 
 class FeatureMapConsolidator(nn.Module):
     def __init__(
@@ -233,14 +274,20 @@ class FeatureMapConsolidator(nn.Module):
 
 # unet
 
+def kernel_and_same_pad(*kernel_size):
+    paddings = tuple(map(lambda k: k // 2, kernel_size))
+    return dict(kernel_size = kernel_size, padding = paddings)
+
 class XUnet(nn.Module):
     def __init__(
         self,
         dim,
         init_dim = None,
         out_dim = None,
+        frame_kernel_size = 1,
         dim_mults = (1, 2, 4, 8),
         num_blocks_per_stage = (2, 2, 2, 2),
+        num_self_attn_per_stage = (0, 0, 0, 1),
         nested_unet_depths = (0, 0, 0, 0),
         nested_unet_dim = 32,
         channels = 3,
@@ -248,15 +295,19 @@ class XUnet(nn.Module):
         resnet_groups = 8,
         consolidate_upsample_fmaps = True,
         skip_scale = 2 ** -0.5,
-        weight_standardize = False
+        weight_standardize = False,
+        attn_heads = 8,
+        attn_dim_head = 32
     ):
         super().__init__()
 
+        self.train_as_images = frame_kernel_size == 1
+
         self.skip_scale = skip_scale
         self.channels = channels
 
         init_dim = default(init_dim, dim)
-        self.init_conv = nn.Conv3d(channels, init_dim, (1, 7, 7), padding = (0, 3, 3))
+        self.init_conv = nn.Conv3d(channels, init_dim, **kernel_and_same_pad(frame_kernel_size, 7, 7))
 
         dims = [init_dim, *map(lambda m: dim * m, dim_mults)]
         in_out = list(zip(dims[:-1], dims[1:]))
@@ -268,7 +319,7 @@ class XUnet(nn.Module):
 
         # resnet or convnext
 
-        blocks = partial(ConvNextBlock) if use_convnext else partial(ResnetBlock, groups = resnet_groups, weight_standardize = weight_standardize)
+        blocks = partial(ConvNextBlock, frame_kernel_size = frame_kernel_size) if use_convnext else partial(ResnetBlock, groups = resnet_groups, weight_standardize = weight_standardize, frame_kernel_size = frame_kernel_size)
 
         # whether to use nested unet, as in unet squared paper
 
@@ -279,6 +330,18 @@ class XUnet(nn.Module):
         num_blocks_per_stage = cast_tuple(num_blocks_per_stage, num_resolutions)
         assert all([num_blocks > 0 for num_blocks in num_blocks_per_stage])
 
+        # number of self attention blocks per stage
+
+        num_self_attn_per_stage = cast_tuple(num_self_attn_per_stage, num_resolutions)
+        assert all([num_self_attn_blocks >= 0 for num_self_attn_blocks in num_self_attn_per_stage])
+
+        # attn kwargs
+
+        attn_kwargs = dict(
+            heads = attn_heads,
+            dim_head = attn_dim_head
+        )
+
         # modules for all layers
 
         skip_dims = []
@@ -286,20 +349,22 @@ class XUnet(nn.Module):
         down_stage_parameters = [
             in_out,
             nested_unet_depths,
-            num_blocks_per_stage
+            num_blocks_per_stage,
+            num_self_attn_per_stage
         ]
 
         up_stage_parameters = [reversed(params[:-1]) for params in down_stage_parameters]
 
         # downs
 
-        for ind, ((dim_in, dim_out), nested_unet_depth, num_blocks) in enumerate(zip(*down_stage_parameters)):
+        for ind, ((dim_in, dim_out), nested_unet_depth, num_blocks, self_attn_blocks) in enumerate(zip(*down_stage_parameters)):
             is_last = ind >= (num_resolutions - 1)
             skip_dims.append(dim_in)
 
             self.downs.append(nn.ModuleList([
                 blocks(dim_in, dim_in, nested_unet_depth = nested_unet_depth, nested_unet_dim = nested_unet_dim),
                 nn.ModuleList([blocks(dim_in, dim_in, nested_unet_depth = nested_unet_depth, nested_unet_dim = nested_unet_dim) for _ in range(num_blocks - 1)]),
+                nn.ModuleList([TransformerBlock(dim_in, depth = self_attn_blocks, **attn_kwargs) for _ in range(self_attn_blocks)]),
                 Downsample(dim_in, dim_out)
             ]))
 
@@ -316,12 +381,13 @@ class XUnet(nn.Module):
 
         # ups
 
-        for ind, ((dim_in, dim_out), nested_unet_depth, num_blocks) in enumerate(zip(*up_stage_parameters)):
+        for ind, ((dim_in, dim_out), nested_unet_depth, num_blocks, self_attn_blocks) in enumerate(zip(*up_stage_parameters)):
             is_last = ind >= (num_resolutions - 1)
 
             self.ups.append(nn.ModuleList([
                 blocks(dim_out + skip_dims.pop(), dim_out, nested_unet_depth = nested_unet_depth, nested_unet_dim = nested_unet_dim),
                 nn.ModuleList([blocks(dim_out, dim_out, nested_unet_depth = nested_unet_depth, nested_unet_dim = nested_unet_dim) for _ in range(num_blocks - 1)]),
+                nn.ModuleList([TransformerBlock(dim_out, depth = self_attn_blocks, **attn_kwargs) for _ in range(self_attn_blocks)]),
                 Upsample(dim_out, dim_in) if not is_last else nn.Identity()
             ]))
 
@@ -342,26 +408,44 @@ class XUnet(nn.Module):
 
         self.final_conv = nn.Sequential(
             blocks(final_dim_in + dim, dim),
-            nn.Conv3d(dim, out_dim, 3, padding = 1)
+            nn.Conv3d(dim, out_dim, **kernel_and_same_pad(frame_kernel_size, 3, 3))
         )
 
     def forward(self, x):
         is_image = x.ndim == 4
+
+        # validations
+
+        assert not (is_image and not self.train_as_images), 'you specified a frame kernel size for the convolutions in this unet, but you are passing in images'
+        assert not (not is_image and self.train_as_images), 'you specified no frame kernel size dimension, yet you are passing in a video. fold the frame dimension into the batch'
+
+        # cast images to 1 framed video
+
         if is_image:
             x = rearrange(x, 'b c h w -> b c 1 h w')
 
+        # initial convolution
+
         x = self.init_conv(x)
+
+        # residual
+
         r = x.clone()
 
+        # downs and ups
+
         down_hiddens = []
         up_hiddens = []
 
-        for init_block, (blocks), downsample in self.downs:
+        for init_block, blocks, attn_blocks, downsample in self.downs:
             x = init_block(x)
 
             for block in blocks:
                 x = block(x)
 
+            for attn_block in attn_blocks:
+                x = attn_block(x)
+
             down_hiddens.append(x)
             x = downsample(x)
 
@@ -373,7 +457,7 @@ class XUnet(nn.Module):
         x = self.mid_upsample(x)
 
 
-        for init_block, blocks, upsample in self.ups:
+        for init_block, blocks, attn_blocks, upsample in self.ups:
             x = torch.cat((x, down_hiddens.pop() * self.skip_scale), dim=1)
 
             x = init_block(x)
@@ -381,12 +465,22 @@ class XUnet(nn.Module):
             for block in blocks:
                 x = block(x)
 
+            for attn_block in attn_blocks:
+                x = attn_block(x)
+
             up_hiddens.insert(0, x)
             x = upsample(x)
 
+        # consolidate feature maps
+
         x = self.consolidator(x, up_hiddens)
 
+        # final residual
+
         x = torch.cat((x, r), dim = 1)
+
+        # final convolution
+
         out = self.final_conv(x)
 
         if is_image:
@@ -406,12 +500,12 @@ class PixelShuffleUpsample(nn.Module):
         super().__init__()
         self.scale_squared = scale_factor ** 2
         dim_out = default(dim_out, dim)
-        conv = nn.Conv2d(dim, dim_out * self.scale_squared, 1)
+        conv = nn.Conv3d(dim, dim_out * self.scale_squared, 1)
 
         self.net = nn.Sequential(
             conv,
             nn.SiLU(),
-            nn.PixelShuffle(scale_factor)
+            Rearrange('b (c r s) f h w -> b c f (h r) (w s)', r = scale_factor, s = scale_factor)
         )
 
         self.init_conv_(conv)
@@ -426,17 +520,7 @@ class PixelShuffleUpsample(nn.Module):
         nn.init.zeros_(conv.bias.data)
 
     def forward(self, x):
-        is_video = x.ndim == 5
-
-        if is_video:
-            f = x.shape[2]
-            x = rearrange(x, 'b c f h w -> (b f) c h w')
-
         x = self.net(x)
-
-        if is_video:
-            x = rearrange(x, '(b f) c h w -> b c f h w', f = f)
-
         return x
 
 class NestedResidualUnet(nn.Module):
@@ -446,6 +530,7 @@ class NestedResidualUnet(nn.Module):
         *,
         depth,
         M = 32,
+        frame_kernel_size = 1,
         add_residual = False,
         groups = 4,
         skip_scale = 2 ** -0.5,
@@ -457,14 +542,14 @@ class NestedResidualUnet(nn.Module):
         self.downs = nn.ModuleList([])
         self.ups = nn.ModuleList([])
 
-        conv = WeightStandardizedConv2d if weight_standardize else nn.Conv2d
+        conv = WeightStandardizedConv3d if weight_standardize else nn.Conv3d
 
         for ind in range(depth):
             is_first = ind == 0
             dim_in = dim if is_first else M
 
             down = nn.Sequential(
-                conv(dim_in, M, 4, stride = 2, padding = 1),
+                conv(dim_in, M, (1, 4, 4), stride = (1, 2, 2), padding = (0, 1, 1)),
                 nn.GroupNorm(groups, M),
                 nn.SiLU()
             )
@@ -479,7 +564,7 @@ class NestedResidualUnet(nn.Module):
             self.ups.append(up)
 
         self.mid = nn.Sequential(
-            conv(M, M, 3, padding = 1),
+            conv(M, M, **kernel_and_same_pad(frame_kernel_size, 3, 3)),
             nn.GroupNorm(groups, M),
             nn.SiLU()
         )
@@ -493,10 +578,6 @@ class NestedResidualUnet(nn.Module):
         if self.add_residual:
             residual = default(residual, x.clone())
 
-        if is_video:
-            f = x.shape[2]
-            x = rearrange(x, 'b c f h w -> (b f) c h w')
-
         *_, h, w = x.shape
 
         assert h == w, 'only works with square images'
@@ -521,9 +602,6 @@ class NestedResidualUnet(nn.Module):
 
         # adding residual
 
-        if is_video:
-            x = rearrange(x, '(b f) c h w -> b c f h w', f = f)
-
         if self.add_residual:
             x = x + residual
             x = F.silu(x)
